from abc import ABC, abstractmethod
from functools import partial

from numpy.typing import NDArray
from scipy.sparse import csc_matrix
from scipy.sparse.linalg import splu


class Solver(ABC):
    @abstractmethod
    def factor(self, m: csc_matrix) -> None:
        ...

    @abstractmethod
    def solve(self, rhs: NDArray, transpose: bool = False) -> NDArray:
        ...

    @abstractmethod
    def clear(self) -> None:
        ...


class SuperLU(Solver):
    _ctx: dict = {}

    def __init__(self, **options):
        self._ctx["splu"] = partial(splu, **options)

    def factor(self, m: csc_matrix) -> None:
        self._ctx["factorization"] = self._ctx["splu"](m)

    def solve(self, rhs: NDArray, transpose: bool = False) -> NDArray:
        return self._ctx["factorization"].solve(rhs, trans="T" if transpose else "N")

    def clear(self) -> None:
        self._ctx.pop("factorization", None)


def get_solver(solver: str) -> Solver:
    match solver:
        case "SuperLU":
            return SuperLU(
                diag_pivot_thresh=0.1,
                permc_spec="MMD_AT_PLUS_A",
                options={"SymmetricMode": True},
            )
        case _:
            raise ValueError(f"Invalid solver: {solver}")



import numpy as np
import pytest
from autograd.test_util import check_grads
from scipy.sparse import coo_matrix

from tofea.primitives import solve_coo
from tofea.solvers import get_solver


@pytest.fixture()
def rng():
    seed = 36523525
    return np.random.default_rng(seed)


@pytest.mark.parametrize("n", [10, 11])
@pytest.mark.parametrize("solver", ["SuperLU"])
@pytest.mark.parametrize("mode", ["fwd", "rev"])
def test_solve_coo_entries_grad(rng, n, solver, mode):
    m = rng.random((n, n))
    m = coo_matrix(m @ m.T)

    b = rng.random(n)

    _solver = get_solver(solver)

    check_grads(
        lambda x: solve_coo(x, (m.row, m.col), b, _solver), modes=[mode], order=1
    )(m.data)


@pytest.mark.parametrize("n", [10, 11])
@pytest.mark.parametrize("solver", ["SuperLU"])
@pytest.mark.parametrize("mode", ["fwd", "rev"])
def test_solve_coo_b_grad(rng, n, solver, mode):
    m = rng.random((n, n))
    m = coo_matrix(m @ m.T)

    b = rng.random(n)

    _solver = get_solver(solver)

    check_grads(
        lambda x: solve_coo(m.data, (m.row, m.col), x, _solver), modes=[mode], order=1
    )(b)
